@@ -10,6 +10,26 @@ struct MTKParameters{T, I, D, C, N, H}
     constant::C
     nonnumeric::N
     caches::H
+
+    function MTKParameters{T, I, D, C, N, H}(tunables::T, initials::I, discrete::D,
+                                             constant::C, nonnumeric::N,
+                                             caches::H) where {T, I, D, C, N, H}
+        if tunables isa StaticVector{0}
+            tunables = SVector{0, eltype(tunables)}()
+        end
+        if initials isa StaticVector{0}
+            initials = SVector{0, eltype(initials)}()
+        end
+        return new{typeof(tunables), typeof(initials), D, C, N, H}(tunables, initials,
+                                                                   discrete, constant,
+                                                                   nonnumeric, caches)
+    end
+    function MTKParameters(tunables::T, initials::I, discrete::D,
+                                             constant::C, nonnumeric::N,
+                                             caches::H) where {T, I, D, C, N, H}
+        return MTKParameters{T, I, D, C, N, H}(tunables, initials, discrete, constant,
+                                               nonnumeric, caches)
+    end
 end
 
 """
@@ -138,11 +158,11 @@ function MTKParameters(
     end
     tunable_buffer = narrow_buffer_type(tunable_buffer; p_constructor)
     if isempty(tunable_buffer)
-        tunable_buffer = SizedVector{0, Float64}()
+        tunable_buffer = SVector{0, Float64}()
     end
     initials_buffer = narrow_buffer_type(initials_buffer; p_constructor)
     if isempty(initials_buffer)
-        initials_buffer = SizedVector{0, Float64}()
+        initials_buffer = SVector{0, Float64}()
     end
     disc_buffer = narrow_buffer_type.(disc_buffer; p_constructor)
     const_buffer = narrow_buffer_type.(const_buffer; p_constructor)
@@ -805,6 +825,10 @@ function Base.getindex(ngi::NestedGetIndex, idx::Tuple)
     i, j, k... = idx
     return ngi.x[i][j][k...]
 end
+function Base.getindex(ngi::NestedGetIndex, idx::NTuple{2})
+    i, j = idx
+    return ngi.x[i][j]
+end
 
 # Required for DiffEqArray constructor to work during interpolation
 Base.size(::NestedGetIndex) = ()
@@ -826,19 +850,27 @@ function SciMLBase.create_parameter_timeseries_collection(
     isempty(ps.discrete) && return nothing
     num_discretes = only(blocksize(ps.discrete[1]))
     buffers = []
-    partition_type = Tuple{(typeof(parent(buf)) for buf in ps.discrete)...}
+    partition_type = typeof(SciMLBase.get_saveable_values(sys, ps, 1))
     for i in 1:num_discretes
         ts = eltype(tspan)[]
-        us = NestedGetIndex{partition_type}[]
+        us = partition_type[]
         push!(buffers, DiffEqArray(us, ts, (1, 1)))
     end
 
     return ParameterTimeseriesCollection(Tuple(buffers), copy(ps))
 end
 
+@inline __get_blocks(tsidx::Int) = ()
+@inline function __get_blocks(tsidx::Int, buffer::BlockedArray, buffers...)
+    (buffer[Block(tsidx)], __get_blocks(tsidx, buffers...)...)
+end
+@inline function __get_blocks(tsidx::Int, buffer::BlockedArray{<:AbstractArray}, buffers...)
+    (copy.(buffer[Block(tsidx)]), __get_blocks(tsidx, buffers...)...)
+end
+
 function SciMLBase.get_saveable_values(
         sys::AbstractSystem, ps::MTKParameters, timeseries_idx)
-    return NestedGetIndex(Tuple(buffer[Block(timeseries_idx)] for buffer in ps.discrete))
+    return NestedGetIndex(__get_blocks(timeseries_idx, ps.discrete...))
 end
 
 function save_callback_discretes!(integ::SciMLBase.DEIntegrator, callback)
@@ -867,10 +899,10 @@ end
 @generated function Base.getindex(
         ps::MTKParameters{T, I, D, C, N, H}, idx::Int) where {T, I, D, C, N, H}
     paths = []
-    if !(T <: SizedVector{0, Float64})
+    if !(T <: StaticVector{0})
         push!(paths, :(ps.tunable))
     end
-    if !(I <: SizedVector{0, Float64})
+    if !(I <: StaticVector{0})
         push!(paths, :(ps.initials))
     end
     for i in 1:fieldcount(D)
@@ -897,10 +929,10 @@ end
 @generated function Base.length(ps::MTKParameters{
         T, I, D, C, N, H}) where {T, I, D, C, N, H}
     len = 0
-    if !(T <: SizedVector{0, Float64})
+    if !(T <: SVector{0})
         len += 1
     end
-    if !(I <: SizedVector{0, Float64})
+    if !(I <: SVector{0})
         len += 1
     end
     len += fieldcount(D) + fieldcount(C) + fieldcount(N) + fieldcount(H)
